from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.builtins import range
from future.builtins import dict
from future.builtins import int
from future import standard_library
standard_library.install_hooks()

import unittest
import familyanalyzer as fa
fa.PROGRESSBAR = False
fa.familyanalyzer.PROGRESSBAR = False


class SetupHelper(object):
    @staticmethod
    def createOrthoXMLParserFromSimpleEx():
        filename = "test/simpleEx.orthoxml"
        return fa.OrthoXMLParser(filename)


class OrthoXMLParserTest(unittest.TestCase):
    def setUp(self):
        self._op = SetupHelper.createOrthoXMLParserFromSimpleEx()

    def test_nrOfToplevelFamilies(self):
        self.assertEqual(len(self._op.getToplevelGroups()), 3)

    def test_returnedSpecies(self):
        expectedSpecies = {'HUMAN', 'PANTR', 'MOUSE', 'RATNO',
                           'CANFA', 'XENTR'}
        self.assertSetEqual(self._op.getSpeciesSet(),
                            expectedSpecies)

    def test_numberOfGenesPerSpecies(self):
        expectedCnts = dict(HUMAN=4, PANTR=4, MOUSE=4, RATNO=2,
                            CANFA=3, XENTR=2)
        allGenes = self._op.getGeneIds()
        for species in expectedCnts.keys():
            self.assertEqual(
                len([gid for gid in allGenes
                    if self._op.mapGeneToSpecies(gid) == species]),
                expectedCnts[species],
                "number of genes not correct for "+species)

    def test_numberOfGenesPerSpecies_specFilter(self):
        expectedCnts = dict(HUMAN=4, PANTR=4, MOUSE=4, RATNO=2,
                            CANFA=3, XENTR=2)
        param_list = [{'HUMAN'}, {'XENTR'}, {'PANTR', 'MOUSE'}, {}]
        for param in param_list:
            expected = sum([expectedCnts[z] for z in param])
            returned = len(self._op.getGeneIds(speciesFilter=param))
            self.assertEqual(expected, returned, 'failed with {}'.format(param))

    def test_xrefMapping(self):
        xreftags = dict(protId='', geneId='g')
        allGenes = self._op.getGeneIds()
        for gid in allGenes:
            species = self._op.mapGeneToSpecies(gid)
            for xreftag, prefix in xreftags.items():
                expectedId = "{}{}{}".format(species, prefix, int(gid) % 10)
                xref = self._op.mapGeneToXRef(gid, xreftag)
                self.assertEqual(xref, expectedId,
                                 "xrefmapping failed for {}: {} vs {}"
                                 .format(gid, xref, expectedId))


class TaxNodeTest(unittest.TestCase):
    def setUp(self):
        self.root = fa.TaxNode('root')
        self.child = fa.TaxNode('child')

        # For the iteration tests
        self.iter_test_root = fa.TaxNode('root')
        self.left = fa.TaxNode('left')
        self.right = fa.TaxNode('right')
        self.leftleft = fa.TaxNode('left-left')
        self.leftright = fa.TaxNode('left-right')
        self.left.add_child(self.leftleft)
        self.left.add_child(self.leftright)
        self.iter_test_root.add_child(self.left)
        self.iter_test_root.add_child(self.right)

    def tearDown(self):
        self.root = None
        self.child = None
        self.iter_test_root = None
        self.left = None
        self.right = None
        self.leftleft = None
        self.leftright = None

    def test_add_child(self):
        self.root.add_child(self.child)
        self.assertIn(self.child, self.root.down)

    def test_nameProperlySet(self):
        self.assertEqual(self.root.name, 'root')

    def test_setParentOnce(self):
        self.child.add_parent(self.root)
        self.assertEqual(self.child.up, self.root)

    def test_setParentTwiceSame(self):
        self.child.add_parent(self.root)
        self.child.add_parent(self.root)
        self.assertTrue(True)

    def test_failWithTwoDifferentParentsSet(self):
        extraNode = fa.TaxNode('extra')
        self.child.add_parent(self.root)
        self.assertRaises(fa.TaxonomyInconsistencyError,
                          self.child.add_parent,
                          extraNode)

    def test_iter_leaves(self):
        leaves = list(self.iter_test_root.iter_leaves())
        self.assertEqual(leaves, [self.leftleft, self.leftright, self.right])

    def test_iter_preorder(self):
        nodes = list(self.iter_test_root.iter_preorder())
        expected = [self.iter_test_root,
                    self.left,
                    self.leftleft,
                    self.leftright,
                    self.right]
        self.assertEqual(nodes, expected)

    def test_iter_postorder(self):
        nodes = list(self.iter_test_root.iter_postorder())
        expected = [self.leftleft,
                    self.leftright,
                    self.left,
                    self.right,
                    self.iter_test_root]
        self.assertEqual(nodes, expected)

    def test_iter_levelorder(self):
        nodes = list(self.iter_test_root.iter_levelorder())
        expected = [self.iter_test_root,
                    self.left,
                    self.right,
                    self.leftleft,
                    self.leftright]
        self.assertEqual(nodes, expected)


class GeneFamilyTest(unittest.TestCase):

    def getLastExampleFamily(self, parser=None):
        if parser is None:
            parser = SetupHelper.createOrthoXMLParserFromSimpleEx()
        fam = parser.getToplevelGroups()[-1]
        return(fam)

    def test_onlyOrthologGroup(self):
        fam = self.getLastExampleFamily()
        paralogNode = fam.find(".//{{{ns0}}}paralogGroup".format(**fa.OrthoXMLParser.ns))
        self.assertRaises(fa.ElementError, fa.GeneFamily, paralogNode)

    def test_members(self):
        fam = self.getLastExampleFamily()
        expectedMembers = {'3', '13', '23', '33', '53', '14', '34'}
        gf = fa.GeneFamily(fam)
        members = set(gf.getMemberGenes())
        self.assertSetEqual(expectedMembers, members)

    def famMemberGenesAtLevel(self, fam, level):
        gf = fa.GeneFamily(fam)
        levelAnalysis = gf.analyzeLevel(level)
        genes = set()
        for subfam in levelAnalysis:
            genes.update(subfam.getMemberGenes())
        return genes

    def test_famMemberAtSubLevel(self):
        parser = SetupHelper.createOrthoXMLParserFromSimpleEx()
        fam = self.getLastExampleFamily(parser)
        cases = {'Primates': {'3', '13'}, 'Euarchontoglires': {'3', '13', '14', '33', '34'}}
        for lev, expMemb in cases.items():
            genes = self.famMemberGenesAtLevel(fam, lev)
            self.assertSetEqual(genes, expMemb)

    def test_simpleAnalyzeStrategy(self):
        """test the classification of genes for a few testcases"""
        parser = SetupHelper.createOrthoXMLParserFromSimpleEx()
        fam = fa.GeneFamily(self.getLastExampleFamily(parser))
        analyzer = fa.BasicLevelAnalysis(parser)
        summary = analyzer.analyzeGeneFam(fam)
        hum = summary['HUMAN']
        self.assertEqual(hum.typ, "SINGLECOPY")
        self.assertSetEqual(hum.genes, {'3'})
        ptr = summary['PANTR']
        self.assertEqual(ptr.typ, "MULTICOPY")
        self.assertSetEqual(ptr.genes, {'13', '14'})


class TwoLineageComparisons(unittest.TestCase):
    """tests the ability to compare the orthoxml at two
    lineages."""
    maxDiff = None

    def getFamHistory(self, parser, level):
        hist = parser.getFamHistory()
        hist.analyzeLevel(level)
        return hist

    def compareLevels(self, lev1, lev2):
        parser = SetupHelper.createOrthoXMLParserFromSimpleEx()
        tax = fa.TaxonomyFactory.newTaxonomy(parser)
        parser.augmentTaxonomyInfo(tax)
        hist1 = self.getFamHistory(parser, lev1)
        hist2 = self.getFamHistory(parser, lev2)
        return hist1.compare(hist2)

    def compareLevelsSingletonAware(self, lev1, lev2):
        parser = SetupHelper.createOrthoXMLParserFromSimpleEx()
        tax = fa.TaxonomyFactory.newTaxonomy(parser)
        parser.augmentTaxonomyInfo(tax)
        parser.augmentSingletons()
        hist1 = self.getFamHistory(parser, lev1)
        hist2 = self.getFamHistory(parser, lev2)
        return hist1.compare(hist2)

    def test_levels(self):
        levPairs = [('Mammalia', 'Primates'),
                    ('Vertebrata', 'Euarchontoglires'),
                    ('Euarchontoglires', 'Rodents'),
                    ('Vertebrata', 'HUMAN'),
                    ('Primates', 'PANTR'),
                    ('Primates', 'HUMAN')]
        expRes = [[fa.FamIdent('1'), fa.FamIdent('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                  [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                  [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', '3.1a')],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamLost('3.1b')]]
        for i in range(len(levPairs)):
            lev1, lev2 = levPairs[i]
            comp = self.compareLevels(lev1, lev2)
            comp.fams.sort(key=lambda x: x.name)
            self.assertListEqual(comp.fams, expRes[i], 'failed for {} vs {}'.format(lev1, lev2))

    def test_summarise(self):
        levPairs = [('Mammalia', 'Primates'),
                    ('Vertebrata', 'Euarchontoglires'),
                    ('Euarchontoglires', 'Rodents'),
                    ('Vertebrata', 'HUMAN'),
                    ('Primates', 'PANTR'),
                    ('Primates', 'HUMAN')]
        expRes = [{u'identical': 2, u'novel': 0, u'lost': 0, u'duplicated': 2, u'singleton': 0},
                  {u'identical': 1, u'novel': 1, u'lost': 0, u'duplicated': 2, u'singleton': 0},
                  {u'identical': 4, u'novel': 0, u'lost': 0, u'duplicated': 0, u'singleton': 0},
                  {u'identical': 1, u'novel': 1, u'lost': 0, u'duplicated': 1, u'singleton': 0},
                  {u'identical': 4, u'novel': 0, u'lost': 0, u'duplicated': 0, u'singleton': 0},
                  {u'identical': 3, u'novel': 0, u'lost': 1, u'duplicated': 0, u'singleton': 0}]
        for i in range(len(levPairs)):
            lev1, lev2 = levPairs[i]
            comp = self.compareLevels(lev1, lev2)
            summary = comp.summarise()
            self.assertDictEqual(summary, expRes[i], 'failed for {} vs {}'.format(lev1, lev2))

    def test_levels_singleton_aware(self):
        levPairs = [('Mammalia', 'Primates'),
                    ('Vertebrata', 'Euarchontoglires'),
                    ('Euarchontoglires', 'Rodents'),
                    ('Vertebrata', 'HUMAN'),
                    ('Primates', 'PANTR'),
                    ('Primates', 'HUMAN')]
        expRes = [[fa.FamIdent('1'), fa.FamIdent('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                  [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                  [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', '3.1a'), fa.FamSingleton('5')],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                  [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamLost('3.1b'), fa.FamSingleton('5')]]
        for i in range(len(levPairs)):
            lev1, lev2 = levPairs[i]
            comp = self.compareLevelsSingletonAware(lev1, lev2)
            comp.fams.sort(key=lambda x: x.name)
            self.assertListEqual(comp.fams, expRes[i], 'failed for {} vs {}'.format(lev1, lev2))

    def test_summarise_singleton_aware(self):
        levPairs = [('Mammalia', 'Primates'),
                    ('Vertebrata', 'Euarchontoglires'),
                    ('Euarchontoglires', 'Rodents'),
                    ('Vertebrata', 'HUMAN'),
                    ('Primates', 'PANTR'),
                    ('Primates', 'HUMAN')]
        expRes = [{u'identical': 2, u'singleton': 0, u'novel': 0, u'lost': 0, u'duplicated': 2},
                  {u'identical': 1, u'singleton': 0, u'novel': 1, u'lost': 0, u'duplicated': 2},
                  {u'identical': 4, u'singleton': 0, u'novel': 0, u'lost': 0, u'duplicated': 0},
                  {u'identical': 1, u'singleton': 1, u'novel': 1, u'lost': 0, u'duplicated': 1},
                  {u'identical': 4, u'singleton': 0, u'novel': 0, u'lost': 0, u'duplicated': 0},
                  {u'identical': 3, u'singleton': 1, u'novel': 0, u'lost': 1, u'duplicated': 0}]
        for i in range(len(levPairs)):
            lev1, lev2 = levPairs[i]
            comp = self.compareLevelsSingletonAware(lev1, lev2)
            summary = comp.summarise()
            self.assertDictEqual(summary, expRes[i], 'failed for {} vs {}'.format(lev1, lev2))

    def test_filter(self):
        levPairs = [('Mammalia', 'Primates'),
                    ('Vertebrata', 'Euarchontoglires'),
                    ('Euarchontoglires', 'Rodents'),
                    ('Vertebrata', 'HUMAN'),
                    ('Primates', 'PANTR'),
                    ('Primates', 'HUMAN')]
        expRes = [
                  [
                   [fa.FamIdent('1'), fa.FamIdent('2')],
                   [fa.FamIdent('1')],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                   [fa.FamIdent('1')],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a')]
                  ],
                  [
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                   [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', ['3.1a', '3.1b'])],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                   [fa.FamIdent('1'), fa.FamNovel('2'), fa.FamDupl('3', '3.1a'), fa.FamSingleton('5')],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamIdent('3.1b')],
                   [fa.FamIdent('1'), fa.FamIdent('2'), fa.FamIdent('3.1a'), fa.FamLost('3.1b'), fa.FamSingleton('5')]
                  ]
                 ]
        filters = ['identical', {'identical', 'lost', 'singleton', 'novel', 'duplicated'}]
        res = [ [], [] ]
        for i in range(len(levPairs)):
            lev1, lev2 = levPairs[i]
            comp = self.compareLevelsSingletonAware(lev1, lev2)
            for j in range(len(filters)):
                fams = comp.filter(filters[j])
                res[j].append(fams)
                self.assertListEqual(res[j][i], expRes[j][i], "failed for {} vs {}".format(lev1, lev2))


class TaxonomyFactoryTest(unittest.TestCase):

    def test_xmlTaxonomyNotImpl(self):
        self.assertRaises(NotImplementedError,
                          fa.TaxonomyFactory.newTaxonomy,
                          "test/taxEx.xml")

    def test_taxFromOrthoXMLParser(self):
        p = SetupHelper.createOrthoXMLParserFromSimpleEx()
        tax = fa.TaxonomyFactory.newTaxonomy(p)
        expectedLevels = p.getLevels().union(p.getSpeciesSet())
        self.assertSetEqual(set(tax.hierarchy.keys()), expectedLevels)


class GeneQueryTest(unittest.TestCase):
    """ tests the OrthoXMLQuery methods getInputGenes and getGroupedGenes,
    with and without species filters
    """

    def setUp(self):
        self._op = SetupHelper.createOrthoXMLParserFromSimpleEx()
        self._tax = fa.TaxonomyFactory.newTaxonomy(self._op)
        self._op.augmentTaxonomyInfo(self._tax)
        self._input_query = fa.OrthoXMLQuery.getInputGenes
        self._grouped_query = fa.OrthoXMLQuery.getGroupedGenes

    def test_all_input_genes(self):
        expected_ids = {'1',  '2',  '3',  '5',  '11', '12', '13', '14',
                        '21', '22', '23', '31', '32', '33', '34', '41',
                        '43', '51', '53'}

        result_nodes = self._input_query(self._op.root)
        result_ids = {n.get('id') for n in result_nodes}
        self.assertEqual(result_ids, expected_ids)

    def test_filtered_input_genes(self):
        filters = ['HUMAN', 'PANTR', 'CANFA', 'MOUSE', 'RATNO', 'XENTR']
        expected_ids = dict(HUMAN={'1',  '2',  '3',  '5'},
                            PANTR={'11', '12', '13', '14'},
                            CANFA={'21', '22', '23'},
                            MOUSE={'31', '32', '33', '34'},
                            RATNO={'41', '43'},
                            XENTR={'51', '53'})

        for filter_ in filters:
            result_nodes = self._input_query(self._op.root, filter_)
            result_ids = {n.get('id') for n in result_nodes}
            self.assertEqual(result_ids, expected_ids[filter_],
                             'failed with {}'.format(filter_))

    def test_all_grouped_genes(self):
        expected_ids = {'1',  '2',  '3',  '11', '12', '13',
                        '14', '21', '22', '23', '31', '32',
                        '33', '34', '41', '51', '53'}

        result_nodes = self._grouped_query(self._op.root)
        result_ids = {n.get('id') for n in result_nodes}
        self.assertEqual(result_ids, expected_ids)

    def test_filtered_grouped_genes(self):
        filters = ['HUMAN', 'PANTR', 'CANFA', 'MOUSE', 'RATNO', 'XENTR']
        expected_ids = dict(HUMAN={'1',  '2',  '3'},
                            PANTR={'11', '12', '13', '14'},
                            CANFA={'21', '22', '23'},
                            MOUSE={'31', '32', '33', '34'},
                            RATNO={'41'},
                            XENTR={'51', '53'})

        for filter_ in filters:
            result_nodes = self._grouped_query(self._op.root, filter_)
            result_ids = {n.get('id') for n in result_nodes}
            self.assertEqual(result_ids, expected_ids[filter_],
                             'failed with {}'.format(filter_))


class TaxonomyNewickTest(unittest.TestCase):
    """Tests that the Taxonomy can write itself as newick """

    def setUp(self):
        self._op = SetupHelper.createOrthoXMLParserFromSimpleEx()
        self._tax = fa.TaxonomyFactory.newTaxonomy(self._op)

    def test_plain_newick(self):
        equivalent_trees = {
            ('((((RATNO, MOUSE)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((MOUSE, RATNO)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((RATNO, MOUSE)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((MOUSE, RATNO)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),

            ('((CANFA, ((RATNO, MOUSE)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((MOUSE, RATNO)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((RATNO, MOUSE)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((MOUSE, RATNO)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),

            ('(XENTR, (((RATNO, MOUSE)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((MOUSE, RATNO)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((RATNO, MOUSE)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((MOUSE, RATNO)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),

            ('(XENTR, (CANFA, ((RATNO, MOUSE)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((MOUSE, RATNO)Rodents, (PANTR, HUMAN)Primates)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((RATNO, MOUSE)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((MOUSE, RATNO)Rodents, (HUMAN, PANTR)Primates)Euarchontoglires)Mammalia)Vertebrata;'),

            ('((((PANTR, HUMAN)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((PANTR, HUMAN)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((HUMAN, PANTR)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),
            ('((((HUMAN, PANTR)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires, CANFA)Mammalia, XENTR)Vertebrata;'),

            ('((CANFA, ((PANTR, HUMAN)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((PANTR, HUMAN)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((HUMAN, PANTR)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),
            ('((CANFA, ((HUMAN, PANTR)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires)Mammalia, XENTR)Vertebrata;'),

            ('(XENTR, (((PANTR, HUMAN)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((PANTR, HUMAN)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((HUMAN, PANTR)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),
            ('(XENTR, (((HUMAN, PANTR)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires, CANFA)Mammalia)Vertebrata;'),

            ('(XENTR, (CANFA, ((PANTR, HUMAN)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((PANTR, HUMAN)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((HUMAN, PANTR)Primates, (RATNO, MOUSE)Rodents)Euarchontoglires)Mammalia)Vertebrata;'),
            ('(XENTR, (CANFA, ((HUMAN, PANTR)Primates, (MOUSE, RATNO)Rodents)Euarchontoglires)Mammalia)Vertebrata;'),
        }

        result = self._tax.newick()
        self.assertIn(result, ' '.join(equivalent_trees))

    def test_decorated_newick(self):
        pass

class GeneTreeNodeTest(unittest.TestCase):

    def setUp(self):
        n = fa.GeneTreeNode('root', 'speciation')
        n.add_child(fa.GeneTreeNode('child1', 'speciation'))
        n.add_child(fa.GeneTreeNode('child2', 'speciation'))
        n.children[0].add_child(fa.GeneTreeNode('grandchild1', 'leaf'))
        n.children[0].add_child(fa.GeneTreeNode('grandchild2', 'leaf'))
        dup = fa.GeneTreeNode('D', 'duplication')
        n.children[1].add_child(dup)
        dup.add_child(fa.GeneTreeNode('grandchild3_A', 'leaf'))
        dup.add_child(fa.GeneTreeNode('grandchild3_B', 'leaf'))
        n.children[1].add_child(fa.GeneTreeNode('grandchild4', 'leaf'))
        self.root = n

    def test_iter(self):
        expected = ('root child1 grandchild1 grandchild2 child2 D'
                    ' grandchild3_A grandchild3_B grandchild4')
        result = ' '.join(x.name for x in self.root)
        self.assertEqual(expected, result)

    def test_delete(self):
        d = self.root.children[1]
        d.delete()
        expected_n = ('root child1 grandchild1 grandchild2 D grandchild3_A'
                      ' grandchild3_B grandchild4')
        expected_d = ('child2')
        result_n = ' '.join(x.name for x in self.root)
        result_d = ' '.join(x.name for x in d)
        self.assertEqual((expected_n, expected_d), (result_n, result_d))

    def test_delete_recursive(self):
        self.setUp()
        d = self.root.children[1]
        d.delete(True)
        expected_n = ('root child1 grandchild1 grandchild2')
        expected_d = ('child2')
        result_n = ' '.join(x.name for x in self.root)
        result_d = ' '.join(x.name for x in d)
        self.assertEqual((expected_n, expected_d), (result_n, result_d))

    def test_reroot(self):
        self.setUp()
        d = self.root.children[1]
        d.reroot()
        expected = ('child2 D grandchild3_A grandchild3_B grandchild4 root'
                    ' child1 grandchild1 grandchild2')
        result = ' '.join(x.name for x in d)
        self.assertEqual(expected, result)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_hooks()

import unittest
import familyanalyzer.taxonomy as tax
import io

class NewickTaxonomyTest(unittest.TestCase):
    def test_binary_tree(self):
        newick_tree = "((Leaf1,Leaf2)Internal1_2,Leaf3)Internal1_3;"
        self.check_root_and_return_rootnode(newick_tree)

    def test_multifurcating_tree(self):
        newick_tree = "((Leaf1,Leaf2)Internal1_2,Leaf3,Leaf4)Internal1_4;"
        root = self.check_root_and_return_rootnode(newick_tree) 
        self.assertEqual(sorted([x.name for x in root.down]), 
                         sorted(['Internal1_2','Leaf3','Leaf4']),
                         'children of root node not as expected')

    def check_root_and_return_rootnode(self, tree):
        fp = io.StringIO(tree)
        taxonomy = tax.NewickTaxonomy(fp)
        root_name = tree[tree.rfind(')')+1:tree.rfind(';')]
        root = taxonomy[root_name]
        self.assertIsNone(root.up, 'not root element')
        self.assertListEqual(list(taxonomy.iterParents(root_name)),[],'not root element')
        return root



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.builtins import super
from future.builtins import next
from future.builtins import chr
from future.builtins import dict
from future.builtins import int
from future.builtins import str
from future import standard_library

standard_library.install_hooks()

# !/usr/bin/env python
#
#  This scripts has the purpose of analyzing an orthoXML file.
#  It does so by providing several methods operating on the file.
#
#                            Adrian Altenhoff, June 2013
#
import lxml.etree as etree
import collections
import itertools
import io
import re
import sys
import logging
from .tools import enum, PROGRESSBAR, setup_progressbar
from .orthoxmlquery import ElementError, OrthoXMLQuery
from .taxonomy import NewickTaxonomy, TaxRangeOrthoXMLTaxonomy, XMLTaxonomy

MAXINT = sys.maxsize
logger = logging.getLogger(__name__)

class OrthoXMLParser(object):
    ns = {"ns0": "http://orthoXML.org/2011/"}  # xml namespace

    def __init__(self, filename):
        """creates a OrthoXMLParser object. the parameter filename
        needs to be a path pointing to the orthoxml file to be
        analyzed."""
        self.doc = etree.parse(filename)
        self.root = self.doc.getroot()
        self.tax = None
        self.singletons = None

        self._buildMappings()  # builds three dictionaries - see def below

    def write(self, filename, **kwargs):
        """Write out the (modified) orthoxml file into a new file.
        kwargs include:
            pretty_print=[True/False],
            xml_declaration=[True/False]
            encoding=[e.g. 'UTF-8']"""
        if 'pretty_print' in kwargs:
            self._remove_whitespace()
        self.doc.write(filename, **kwargs)

    def _remove_whitespace(self):
        """Remove whitespace from text and tail fields so that lxml can
        pretty_print the tree"""
        for element in self.root.iter():
            if element.text is not None and element.text.isspace():
                element.text = None
            element.tail = None

    def mapGeneToXRef(self, id_, typ='protId'):
        """
        Looks up id_ (integer id_ number as a string) and type in self._xrefs
        dictionary (aka self._xrefs)
        Returns lookup value, which is a gene name
        """
        if typ is None:
            res = id_
        else:
            tup = (id_, typ)
            res = self._xrefs.get(tup, None)
            if res is None:
                # fallback, if not in dict
                gene = OrthoXMLQuery.getGeneFromId(id_, self.root)
                # return desired ID typ, otherwise whole element
                if typ is not None:
                    res = gene.get(typ, gene)
                else:
                    res = gene
        return res

    def getSpeciesSet(self):
        return self._species  # all species in the xml tree

    def getGeneIds(self, speciesFilter=None, tag="name"):
        genes = list(self._gene2species.keys())
        if speciesFilter is not None:
            genes = [g for g in genes
                     if self._gene2species[g].get(tag, None) in speciesFilter]
        return genes

    def getToplevelGroups(self):
        """A function yielding the toplevel orthologGroups from the file.
        This corresponds to gene families for the purpose of this project."""
        return OrthoXMLQuery.getToplevelOrthologGroups(self.root)

    def getSubFamilies(self, level, root=None):
        """return a forest of orthologGroup nodes with roots at the
        given taxonomic level. This function requires that the
        orthologGroup nodes are annotated with a 'property' element
        with 'TaxRange' as a name and the actual level in a 'value'
        attribute. This is not required by the orthoxml schema."""
        if root is None:
            root = self.root

        return OrthoXMLQuery.getGroupsAtLevel(level, root)

    def get_species_below_node(self, node, return_gene_total_count=False, species_as_name=True):
        """ return a set of all species that have a geneRef present beneath
        the specified node

        :param node: Node of interest
        :param return_gene_total_count: boolean flag whether or not to return
               the number of genes below this node.
        :return:
        """
        generef_nodes = OrthoXMLQuery.getGeneRefNodes(node)
        elem = None if not species_as_name else "name"
        species_covered = {self.mapGeneToSpecies(gr.get('id'), typ=elem)
                           for gr in generef_nodes}
        if return_gene_total_count:
            return species_covered, len(generef_nodes)
        else:
            return species_covered

    @classmethod
    def is_ortholog_group(cls, element):
        """
        Returns true if the passed element is an orthologGroup xml node
        """
        return element.tag == '{{{ns0}}}orthologGroup'.format(**cls.ns)

    @classmethod
    def is_paralog_group(cls, element):
        """
        Returns true if the passed element is an paralogGroup xml node
        """
        return element.tag == '{{{ns0}}}paralogGroup'.format(**cls.ns)

    @classmethod
    def is_evolutionary_node(cls, element):
        """Returns true if the passed element is an evolutionary event
        xml node, i.e. if it is either an orthologGroup or a
        paralogGroup element."""
        return (cls.is_ortholog_group(element) or
                cls.is_paralog_group(element))

    def mapGeneToSpecies(self, id_, typ='name'):
        """
        Does a lookup in the self._gene2species dict:
        key = idnum, return = species name
        """
        if self._gene2species is None:
            self._buildMappings()
        if typ is None:
            return self._gene2species[id_]
        return self._gene2species[id_].get(typ)

    def _findSubNodes(self, targetNode, root=None):
        """return all (recursively) found elements with tagname
        'targetNode' below the root element. If no root element
        is provided, search starts at the document root."""
        rootNode = root if root is not None else self.root
        return rootNode.findall(".//{{{0}}}{1}".
                                format(self.ns['ns0'], targetNode))

    def _buildMappings(self):
        """Builds two dictionaries:
          self._gene2species - keys are ID numbers, values are species
          self._xrefs - keys are tuples
              (idnum, idtype ['geneId','protId']),
          values are gene names
        Also builds the set:
          self._species - All species names in the xml tree"""
        mapping = dict()
        xref = dict()
        for species in self._findSubNodes("species"):
            genes = self._findSubNodes("gene", root=species)
            for gene in genes:
                id_ = gene.get('id')
                mapping[id_] = species
                for tag in gene.keys():
                    if tag != "id":
                        xref[(id_, tag)] = gene.get(tag)
        self._gene2species = mapping
        self._xrefs = xref
        self._species = frozenset({z.get('name') for z in mapping.values()})
        self._levels = frozenset({n.get('value')
                                  for n in self._findSubNodes("property")
                                  if n.get('name') == "TaxRange"})

    def getUbiquitusFamilies(self, minCoverage=.5):
        families = self.getToplevelGroups()
        return [x for x in families if len(self.getGenesPerSpeciesInFam(x)) >=
                minCoverage * len(self.getSpeciesSet())]

    def getLevels(self):
        return self._levels

    def getGenesPerSpeciesInFam(self, fam):
        """
        Takes a gene family, returns a dictionary:
          keys   = species names
          values = set of geneIds belonging to that species at a level
                descended from the family
        """
        genes = collections.defaultdict(set)
        geneRefs = self._findSubNodes("geneRef", fam)
        for gref in geneRefs:
            gid = gref.get('id')
            sp = self.mapGeneToSpecies(gid)
            genes[sp].add(gid)
        return genes

    def getFamHistory(self):
        """This method returns a FamHistory object initialized with
        the most powerful LevelAnalysis handler. The handler depends
        on whether the parser contains a tax attribute, set by the
        augmentTaxonomyInfo method if called."""

        # assure that orthologGroup xml elements annotated with an 'og' attr
        if self.root.find(".//*[@og]") is None:
            GroupAnnotator(self).annotateDoc()

        analyzer = LevelAnalysisFactory().newLevelAnalysis(self)
        return FamHistory(self, analyzer)

    def augmentTaxonomyInfo(self, tax, propagate_top=False):
        """Assign a taxonomy to the orthoxml file. this taxonomy
        is used to augment the xml with the relevant level infos
        as 'TaxRange' property tags in orthologGroup elements.

        The propagate_top parameter can be should be used to enable
        or disable the propagation of all the levels which are
        older than the families topmost level. In other words, if
        enabled, all families arouse at LUCA, otherwise families
        can be invented later on in evolution."""
        if self.tax is not None:
            raise Exception("a taxonomy can be assigned only once")
        self.tax = tax
        GroupAnnotator(self).annotateMissingTaxRanges(tax, propagate_top)

    def augmentSingletons(self):
        GroupAnnotator(self).annotateSingletons()


class TaxonomyFactory(object):
    @classmethod
    def newTaxonomy(cls, arg):
        if isinstance(arg, str):
            if arg.endswith('.xml'):
                return XMLTaxonomy(arg)
            else:
                suffix = arg[arg.rindex('.'):]
                if suffix in ['.nwk', '.tree', '.newick']:
                    return NewickTaxonomy(arg)
        elif isinstance(arg, OrthoXMLParser):
            return TaxRangeOrthoXMLTaxonomy(arg)
        else:
            raise NotImplementedError("unknown type of Taxonomy")


class GeneFamily(object):
    """GeneFamily(root_element)

    Represents one gene family rooted at an orthologous group. """

    def __init__(self, root_element):
        if not OrthoXMLParser.is_ortholog_group(root_element):
            raise ElementError('Not an orthologGroup node')
        self.root = root_element

    def __repr__(self):
        return '{} (id#={})'.format(self.__class__.__name__, self.getFamId())

    def __lt__(self, other):
        return self._cmp() < other._cmp()

    def __gt__(self, other):
        return self._cmp() > other._cmp()

    def __le__(self, other):
        return self._cmp() <= other._cmp()

    def __ge__(self, other):
        return self._cmp() >= other._cmp()

    def __eq__(self, other):
        return self._cmp() == other._cmp()

    def __str__(self):
        return self.root.get('og')

    def prefix_match(self, other):
        query = self._cmp()
        target = other._cmp()
        return target[:len(query)] == query

    def getMemberGenes(self):
        """
        Get all genes belonging to this family. Cached to speed up repeated calls.
        """
        if hasattr(self, '_member_genes'):
            return self._member_genes
        else:
            members = self.root.findall('.//{{{ns0}}}geneRef'.
                                        format(**OrthoXMLParser.ns))
            self._member_genes = [x.get('id') for x in members]
            return self._member_genes

    def getFamId(self):
        return self.root.get('og')

    def _cmp(self):
        """
        Enable sorting based on LOFT numbering scheme,
        such that (e.g.)
            1.1a < 1.1b,
            1.1b < 1.1c,
            1.1z < 1.1aa,
            2 < 10,
            10 < n/a,
        """
        fam = self.getFamId()
        comp = tuple((0, int(num),) if num else (len(alpha.strip('.')),
                                                 alpha.strip('.'),) for (num, alpha) in re.findall(r'(\d+)|(\D+)',
                                                                                                   fam))
        return comp

    def getLevels(self):
        return OrthoXMLQuery.getLevels(self.root)

    def analyzeLevel(self, level):
        """analyze the structure of the family at a given taxonomic
        level.
        returns a list of GeneFamily object, one per sub-family"""

        subFamNodes = OrthoXMLQuery.getGroupsAtLevel(level, self.root)
        subFams = [GeneFamily(fam) for fam in subFamNodes]
        return subFams

    def analyze(self, strategy, level):
        """analyze the history of the GeneFamily using the strategy
        passed to the method. The strategy arguement must be an
        object providing a analyzeGeneFam method,
        e.g. a LevelAnalysis object."""
        self.summary = strategy.analyzeGeneFam(self, level)

    def write(self, fd, speciesFilter=None, idFormatter=lambda x: x):
        species = list(self.summary.keys())
        if speciesFilter is not None:
            species = [g for g in species if g in speciesFilter]
        for spec in self.summary.keys():
            if not spec in species:
                continue
            sumElem = self.summary[spec]
            refs = "; ".join([idFormatter(gid) for gid in sumElem.genes])
            fd.write("{}\t{}\t{}\t{}:{}\n".format(
                self.getFamId(), spec, len(sumElem.genes),
                sumElem.typ, refs))

    def is_singleton(self):
        if not hasattr(self, 'summary'):
            return False
        return 'SINGLETON' in {x.typ for x in self.summary.values()}


class Singletons(GeneFamily):
    def __init__(self, element):
        memb = set()
        if isinstance(element, str):
            memb.add(element)
        else:
            memb.update(element)
        self.members = memb

    def getMemberGenes(self):
        return self.members

    def getFamId(self):
        return "n/a"

    def getLevels(self):
        return None

    def analyzeLevel(self, level):
        return self

    def analyze(self, strategy):
        super().analyze(strategy)
        for sumElement in self.summary.values():
            sumElement.typ = "SINGLETON"


class SummaryOfSpecies(object):
    def __init__(self, typ, genes):
        self.typ = typ
        self.genes = genes


class LevelAnalysisFactory(object):
    def __init__(self):
        pass

    def newLevelAnalysis(self, parser):
        """return a the appropriate LevelAnalysis instance based on
        the presents/absence of a taxonomy in the parser"""
        if parser.tax is None:
            return BasicLevelAnalysis(parser)
        elif parser.singletons is None:
            return TaxAwareLevelAnalysis(parser, parser.tax)
        else:
            return SingletonAwareLevelAnalysis(parser,
                                               parser.tax,
                                               parser.singletons)


class BasicLevelAnalysis(object):
    GeneClasses = enum("MULTICOPY",
                       "SINGLECOPY",
                       "ANCIENT_BUT_LOST",
                       "LATER_GAINED",
                       "SINGLETON")

    def __init__(self, parser):
        self.parser = parser

    def analyzeGeneFam(self, fam, level=None):
        """analyzes a single gene family and returns a summary dict.

        This method classifies all genes in the family depending on
        the number of copies per genome into MULTICOPY or SINGLECOPY
        genes."""

        spec2genes = collections.defaultdict(set)
        for geneId in fam.getMemberGenes():
            spec = self.parser.mapGeneToSpecies(geneId)
            spec2genes[spec].add(geneId)
        summary = dict()
        for spec in iter(spec2genes.keys()):
            nrMemb = len(spec2genes[spec])
            gclass = (self.GeneClasses.MULTICOPY
                      if nrMemb > 1
                      else self.GeneClasses.SINGLECOPY)

            summary[spec] = SummaryOfSpecies(self.GeneClasses.reverse[gclass],
                                             spec2genes[spec])
        return summary


class TaxAwareLevelAnalysis(BasicLevelAnalysis):
    def __init__(self, parser, tax):
        super().__init__(parser)
        self.tax = tax

    def addLosses(self, fam, summary, level):
        lev = fam.getLevels()
        if lev is not None:
            # if several levels exist at this node, use oldest one
            # that is not older than `level'
            lev = self.tax.younger_than_filter(lev, level)
            mostGeneralLevel = self.tax.mostGeneralLevel(lev)
            speciesCoveredByLevel = self.tax.descendents[mostGeneralLevel]
            lostSpecies = speciesCoveredByLevel.difference(summary.keys())
            for lost in lostSpecies:
                summary[lost] = SummaryOfSpecies("ANCIENT_BUT_LOST", [])

    def analyzeGeneFam(self, fam, level):
        """analyzes a single gene family in the context of a known
        taxonomic tree.

        in addition to the method defined in the base class, this
        method adds information of lost genes. It does this by
        checking whether a species within the taxonomic range of
        the family contains a copy of the gene. if not, it had
        been lost."""
        summary = super().analyzeGeneFam(fam, level)
        self.addLosses(fam, summary, level)
        return summary


class SingletonAwareLevelAnalysis(TaxAwareLevelAnalysis):
    def __init__(self, parser, tax, singletons):
        super().__init__(parser, tax)
        self.singletons = singletons

    def analyzeGeneFam(self, fam, level):
        """analyzes a single gene family and returns a summary dict.

        This method classifies all genes in the family depending on
        the number of copies per genome into MULTICOPY or SINGLECOPY
        genes."""

        spec2genes = collections.defaultdict(set)
        for geneId in fam.getMemberGenes():
            spec = self.parser.mapGeneToSpecies(geneId)
            spec2genes[spec].add(geneId)
        summary = dict()
        famId = fam.getFamId()
        for spec, set_ in spec2genes.items():
            if len(set_) > 1:
                gclass = self.GeneClasses.MULTICOPY
            else:
                if famId in self.parser.singletons:
                    gclass = self.GeneClasses.SINGLETON
                else:
                    gclass = self.GeneClasses.SINGLECOPY

            summary[spec] = SummaryOfSpecies(self.GeneClasses.reverse[gclass],
                                             set_)

        self.addLosses(fam, summary, level)
        return summary


class FamHistory(object):
    XRefTag = None

    def __init__(self, parser, analyzer):
        self.parser = parser
        self.analyzer = analyzer
        self._geneFamList = list()

    def __len__(self):
        return self.get_number_of_fams(singletons=True)

    def __getitem__(self, key):
        return self.geneFamDict[key]

    def __iter__(self):
        return iter(self.geneFamList)

    @property
    def geneFamList(self):
        return self._geneFamList

    @geneFamList.setter
    def geneFamList(self, gfam_list):
        self._geneFamList = sorted(gfam_list)

    def setXRefTag(self, tag):
        """set the attribute name of the 'gene' elements which should
        be used for report. defined by orthoxml are 'geneId' and
        'protId'. If not defined, the (numerical) internal ids are used."""
        self.XRefTag = tag

    def analyzeLevel(self, level):
        subFamNodes = OrthoXMLQuery.getGroupsAtLevel(level, self.parser.root)
        gfamList = [GeneFamily(fam) for fam in subFamNodes]

        for gfam in gfamList:
            gfam.analyze(self.analyzer, level)

        self.geneFamDict = {gf.getFamId(): gf for gf in gfamList}
        self.geneFamList = gfamList
        self.analyzedLevel = level

    def write(self, fd, speciesFilter=None):
        """writes the FamHistory object to a given stream object
        in a human readable format.
        The stream object needs to have a write(str) method defined.
        The optional speciesFilter argument accepts a set of
        species names for which the genes in the families are
        reported."""

        formatter = lambda gid: self.parser.mapGeneToXRef(gid, self.XRefTag)
        fd.write("FamilyAnalysis at {}\n".format(self.analyzedLevel))
        for fam in self:
            fam.write(fd, speciesFilter, idFormatter=formatter)

    def __str__(self):
        fd = io.StringIO()
        self.write(fd)
        res = fd.getvalue()
        fd.close()
        return res

    def compare(self, other):
        """ compares two FamilyHistory objects in linear time
            algorithm implemented in Comparer class
        """
        c = Comparer(self, other)
        c.run()
        return c.comp

    def get_number_of_fams(self, singletons=False):
        if singletons:
            return len(self.geneFamList)
        return len([x for x in self if not x.is_singleton()])


class Comparer(object):
    """
    Compares two FamilyHistory objects in linear time.

    Algorithm:

    Init:
    preprocess input by sorting FamilyHistory geneFamLists
    maintain pointers to the lists that only move forwards

    Run:
    If families exactly match:
        Mark as FamIdent
        advance both lists one element
    Else if family2 is a prefix match of family 1:
        While family1.prefix_match(family2):
            Mark as FamDupl
            advance list 2
        advance list 1
    Else if family 1 < family 2 (according to sort comparison)
        While family 1 < family 2:
            Mark as FamLost
            advance list 1
    Else if family 1 > family 2
        While family 1 > family 2:
            Mark as FamNovel
            advance list 2
    When list 1 is exhausted:
        mark all remaining members of list 2 as FamNovel
    When list 2 is exhausted:
        mark all remaining members of list 1 as FamLost

    End:
    When both lists are exhausted
    """

    def __init__(self, fam_history_1, fam_history_2):
        self.i1 = iter(fam_history_1.geneFamList)
        self.i2 = iter(fam_history_2.geneFamList)
        self.f1 = None
        self.f2 = None
        self.advance_i1()
        self.advance_i2()
        self.comp = LevelComparisonResult(fam_history_1.analyzedLevel,
                                          fam_history_2.analyzedLevel)

    def run(self):
        while self.f1 is not None and self.f2 is not None:
            if self.f1 == self.f2:
                self.ident()

            elif self.f1.prefix_match(self.f2):
                self.dupl()

            elif self.f1 < self.f2:
                self.lost()

            elif self.f1 > self.f2:
                self.novel()

            else:
                raise Exception('Unexpected state')

        if self.f1 is None:
            self.l1_exhausted()

        if self.f2 is None:
            self.l2_exhausted()

    def ident(self):
        self.comp.addFamily(FamIdent(self.f1.getFamId()))
        self.advance_i1()
        self.advance_i2()

    def dupl(self):
        m = list()
        while self.f1.prefix_match(self.f2):
            m.append(self.f2)
            self.advance_i2()
            if self.f2 is None:
                break
        self.comp.addFamily(FamDupl(self.f1, m))
        self.advance_i1()

    def lost(self):
        while self.f1 < self.f2 and not self.f1.prefix_match(self.f2):
            self.comp.addFamily(FamLost(self.f1.getFamId()))
            self.advance_i1()
            if self.f1 is None:
                break

    def novel(self):
        while self.f1 > self.f2 and not self.f1.prefix_match(self.f2):
            _Event = (FamSingleton if self.f2.is_singleton()
                      else FamNovel)  # this check is probably redundant
            # because there shouldn't be any
            # singletons if l1 is not exhausted
            # (singletons are always annotated last)
            self.comp.addFamily(_Event(self.f2.getFamId()))
            self.advance_i2()
            if self.f2 is None:
                break

    def advance_i1(self):
        try:
            val = next(self.i1)
        except StopIteration:
            val = None
        self.f1 = val

    def advance_i2(self):
        try:
            val = next(self.i2)
        except StopIteration:
            val = None
        self.f2 = val

    def l1_exhausted(self):
        while self.f2 is not None:
            _Event = (FamSingleton if self.f2.is_singleton()
                      else FamNovel)
            self.comp.addFamily(_Event(self.f2.getFamId()))
            self.advance_i2()

    def l2_exhausted(self):
        while self.f1 is not None:
            self.comp.addFamily(FamLost(self.f1.getFamId()))
            self.advance_i1()


class FamEvent(object):
    event = None

    def __init__(self, fam):
        self.name = str(fam)
        self.fam = fam

    def __str__(self):
        return "{}: {}\n".format(self.name, self.event)

    def __eq__(self, other):
        return self.name == other.name and self.event == other.event

    def __repr__(self):
        return "{}: {}".format(self.name, self.event)


class FamIdent(FamEvent):
    event = "identical"


class FamNovel(FamEvent):
    event = "novel"


class FamLost(FamEvent):
    event = "lost"


class FamSingleton(FamEvent):
    """A single-member 'family' consisting of a gene that doesn't
    match any other family. Only occurs in a leaf."""
    event = "singleton"


class FamDupl(FamEvent):
    event = "duplicated"

    def __init__(self, fam, subfam):
        super().__init__(fam)
        if isinstance(subfam, list):
            subfam_names = "; ".join([str(s) for s in subfam])
        else:
            subfam_names = str(subfam)
        self.into = subfam_names
        self.subfams = subfam

    def __str__(self):
        return self.write()

    def __eq__(self, other):
        return super().__eq__(other) and self.into == other.into

    def write(self):
        ''' Construct output for printing. Shows the ids of the duplicated
            GeneFamily and the resulting subfamilies. It also lists the
            members of the original GeneFamily and the new subfamilies.
            Sample output:
            -------
            114 --> 114.2a; 114.2b
            114: 47953; 28082; 11418; 29862; 117; 50097; 13634; 50845; 14300
            114.2a: 50097; 13634
            114.2b: 50845; 14300
            -------
        '''
        output = "-------\n"
        output += ("{} --> {}\n".format(self.fam.getFamId(), self.into))
        members = "; ".join(self.fam.getMemberGenes())
        output += ("{}: {}\n".format(self.fam.getFamId(), members))
        for group in self.subfams:
            members = "; ".join(group.getMemberGenes())
            output += ("{}: {}\n".format(group.getFamId(), members))
        output += ("-------\n")
        return output


class LevelComparisonResult(object):
    """
    Result of comparing two 'FamHistory' objects at different levels
    on the tree.

    """

    groups = {'identical': 0,
              'duplicated': 1,
              'lost': 2,
              'novel': 3,
              'singleton': 4}

    groups_back = {0: 'identical',
                   1: 'duplicated',
                   2: 'lost',
                   3: 'novel',
                   4: 'singleton'}

    @staticmethod
    def sort_key(item):

        if item.name == 'n/a':
            return (MAXINT,)

        return tuple((int(num) if num else alpha) for
                     (num, alpha) in re.findall(r'(\d+)|(\D+)', item.name))

    def group_sort_key(self, item):
        return tuple(itertools.chain((self.group_key(item),),
                                     self.sort_key(item)))

    def group_key(self, item):
        return self.groups[item.event]

    def __init__(self, lev1, lev2):
        self.fams_dict = dict()
        self.lev1 = lev1
        self.lev2 = lev2

    def __str__(self):
        fd = io.StringIO()
        self.write(fd)
        res = fd.getvalue()
        fd.close()
        return res

    def __getitem__(self, key):
        return self.fams_dict[key]

    def __iter__(self):
        return iter(self.fams)

    @property
    def fams(self):
        return sorted(self.fams_dict.values(), key=self.sort_key)

    def addFamily(self, famEvent):
        self.fams_dict[famEvent.name] = famEvent

    def write(self, fd):
        fd.write("\nLevelComparisonResult between taxlevel {} and {}\n".
                 format(self.lev1, self.lev2))
        for fam in self:
            fd.writelines(str(fam))

    def summarise(self):
        summary = {'identical': 0,
                   'novel': 0,
                   'lost': 0,
                   'duplicated': 0,
                   'singleton': 0}

        for family in self.fams_dict.values():
            if family.event in ['identical', 'lost', 'singleton', 'novel']:
                summary[family.event] += 1
            elif family.event == 'duplicated':
                summary['duplicated'] += len(family.into.split('; '))

        self.summary = summary
        return summary

    def filter(self, filters):
        if not isinstance(filters, set):
            filters = {filters}
        if not filters.issubset({'identical', 'lost', 'singleton', 'novel',
                                 'duplicated'}):
            raise Exception('Unexpected filters: {0}'.format(filters))
        return [x for x in self if x.event in filters]

    def group_fams(self):
        return dict([(self.groups_back[num], list(iterator))
                     for (num, iterator)
                     in itertools.groupby(sorted(self.fams_dict.values(),
                                                 key=self.group_sort_key),
                                          self.group_key)])


class GroupAnnotator(object):
    """this class annotates orthologGroup elements with the numbering
    schema presented in the LOFT paper:
    van der Heijden, Snel, van Noort, Huynen
    Orthology prediction at scalable resolution by phylogenetic tree analysis.
    BMC Bioinformatics, 2007, 8, 83

    and adding additional property tags with skipped TaxRange levels."""

    def __init__(self, parser, taxrange_2_taxid=None, taxid_2_code=None):
        self.parser = parser
        self.ns = parser.ns
        self.taxrange_2_taxid = {rng: str(val) for rng, val in taxrange_2_taxid.items()} if taxrange_2_taxid is not None else {}
        self.taxid_2_code = {k: v for k, v in taxid_2_code.items()} if taxid_2_code is not None else {}

    def _getNextSubId(self, idx):
        """helper method to return the next number at a given depth of
        duplication (idx)"""
        while len(self.dupCnt) < idx:
            self.dupCnt.append(0)
        self.dupCnt[idx - 1] += 1
        return self.dupCnt[idx - 1]

    def _encodeParalogClusterId(self, prefix, nr):
        """encode the paralogGroups at the same level, e.g. 1a, 1b, 1c
        for 3 paralogGroups next to each other. the nr argument
        identifies the individual indices of those 3 paralogGroups."""
        letters = []
        while nr // 26 > 0:
            letters.append(chr(97 + (nr % 26)))
            nr = nr // 26 - 1
        letters.append(chr(97 + (nr % 26)))
        return prefix + ''.join(letters[::-1])  # letters were in reverse order

    def _annotateGroupR(self, node, og, idx=0):
        """create the og attributes at the orthologGroup elements
        according to the naming schema of LOFT. ParalogGroup elements
        do not get own attributes (not possible in the xml schema),
        but propagate their sub-names for the subsequent orthologGroup
        elements."""
        if self.parser.is_ortholog_group(node):
            taxid_node = OrthoXMLQuery.getTaxidNodes(node, recursively=False)
            if len(taxid_node) > 0:
                node.set('id', '{}_{}'.format(og, taxid_node[0].get('value')))
            else:
                node.set('og', og)
            for child in list(node):
                self._annotateGroupR(child, og, idx)
        elif self.parser.is_paralog_group(node):
            idx += 1
            nextOG = "{}.{}".format(og, self._getNextSubId(idx))
            for i, child in enumerate(list(node)):
                self._annotateGroupR(child,
                                     self._encodeParalogClusterId(nextOG, i),
                                     idx)

    def _map_species_element_to_name_and_code(self, species_elem):
        try:
            code = self.taxid_2_code[int(species_elem.get('NCBITaxId'))]
        except KeyError:
            code = None
        name_by_name = species_elem.get('name')
        return name_by_name, code

    def _addTaxRangeR(self, node, noUpwardLevels=False):
        """recursive method to add TaxRange property tags."""
        if self.parser.is_ortholog_group(node) or self.parser.is_paralog_group(node) \
                or OrthoXMLQuery.is_geneRef_node(node):
            species_covered, nr_genes = self.parser.get_species_below_node(node, return_gene_total_count=True, species_as_name=False)
            species_covered = set(self.tax.map_potential_internal_speciesname_to_leaf(
                *self._map_species_element_to_name_and_code(s))
                                  for s in species_covered)
            current_level = self.tax.mrca(species_covered)
            og_tag = '{{{}}}orthologGroup'.format(OrthoXMLQuery.ns['ns0'])

            if self.parser.is_ortholog_group(node):
                comp_score = OrthoXMLQuery.getScoreNodes(node, 'CompletenessScore')
                if len(comp_score) == 0:
                    node.append(self._createCompletnessScoreTag(current_level, species_covered))
                node.append(self._createNrMemberGeneTag(nr_genes))
                taxrange = OrthoXMLQuery.getTaxRangeNodes(node, False)
                taxid = OrthoXMLQuery.getTaxidNodes(node, False)
                if len(taxrange) > 0:
                    # check consistency between current_level and value stored in taxrange
                    if taxrange[0].get('value') != current_level:
                        raise Exception("Inconsistent TaxRange: {} vs current_level {}"
                                        .format(taxrange[0].get('value'), current_level))
                    if len(taxid) > 0:
                        if taxid[0].get('value') != self.taxrange_2_taxid[current_level]:
                            raise Exception("Inconsitency between taxids: {} vs {}"
                                            .format(taxid[0].get('value'), self.taxrange_2_taxid[current_level]))
                    else:
                        try:
                            node.append(self._create_taxid(current_level))
                        except KeyError:
                            pass
                else:
                    node.append(self._createTaxRangeTags(current_level))

            try:  # find the closest ancestral orthogroup if it has a TaxRange property
                parent_orthogroup = next(node.iterancestors(og_tag))
                parent_levels = {z.get('value')
                                 for z in OrthoXMLQuery.getTaxRangeNodes(parent_orthogroup, False)}
            except StopIteration:  # couldn't find a parent with a TaxRange property; no extra annotation possible
                parent_levels = set([])

            if len(parent_levels) > 0:
                most_recent_parent_level = self.tax.mostSpecific(parent_levels)

                # Ortholog Node - append missing tax range(s) as property tags under the current node
                if self.parser.is_ortholog_group(node):
                    self._insertOGs_between(node.getparent(), node, current_level, most_recent_parent_level, nr_genes,
                                            species_covered, include_self=False)

                # Paralog Node - insert ortholog node between self and parent; add missing tax range(s) to new parent
                elif self.parser.is_paralog_group(node):
                    if self.tax.levels_between(most_recent_parent_level, current_level) > 1:
                        self._insertOGs_between(node.getparent(), node, current_level, most_recent_parent_level,
                                                nr_genes, species_covered, include_self=False)

                # GeneRef Node - insert ortholog node between self and parent; add all tax range(s) to new parent
                else:
                    self._insertOGs_between(node.getparent(), node, current_level, most_recent_parent_level,
                                            nr_genes, species_covered, include_self=True)
                    return

            for child in node:
                self._addTaxRangeR(child, noUpwardLevels)

    def _insertOGs_between(self, parent, child, specificLev, beforeLev, nr_genes, covered_species, include_self=True):
        pos = parent.index(child)
        if include_self:
            child = self._insert_one_OG(child, specificLev, covered_species, nr_genes)
        for lev in self.tax.iterParents(specificLev, stopBefore=beforeLev):
            child = self._insert_one_OG(child, lev, covered_species=covered_species, nr_genes=nr_genes)
        parent.insert(pos, child)

    def _insert_one_OG(self, child, level, covered_species, nr_genes):
        el = etree.Element('{{{ns0}}}orthologGroup'.format(**self.parser.ns))
        el.append(self._createCompletnessScoreTag(level, covered_species))
        el.extend(self._createTaxRangeTags(level))
        el.append(self._createNrMemberGeneTag(nr_genes))
        el.append(child)
        return el

    def _createTaxRangeTags(self, lev, **kwargs):
        tags = [self._create_tax_range(lev, **kwargs)]
        try:
            tags.append(self._create_taxid(lev))
        except KeyError:
            pass
        return tags

    def _create_tax_range(self, lev, **kwargs):
        return etree.Element('{{{ns0}}}property'.format(**self.parser.ns),
                             attrib=dict(name='TaxRange', value=lev, **kwargs))

    def _create_taxid(self, lev):
        return etree.Element('{{{ns0}}}property'.format(**self.parser.ns),
                             attrib=dict(name='taxid',
                                         value=self.taxrange_2_taxid[lev]))

    def _createNrMemberGeneTag(self, nr_genes):
        return etree.Element('{{{ns0}}}property'.format(**self.parser.ns),
                             attrib={"name": "NrMemberGenes", "value": str(nr_genes)})

    def completenessScore(self, level, covered_species):
        return "{:.3f}".format(len(covered_species) / len(self.tax.descendents[level]))

    def _createCompletnessScoreTag(self, level, covered_species):
        el = etree.Element('{{{ns0}}}score'.format(**self.parser.ns),
                           attrib={"id": "CompletenessScore",
                                   "value": self.completenessScore(level, covered_species)})
        return el

    def annotateMissingTaxRanges(self, tax, propagate_top=False, verbosity=0):
        """This function adds left-out taxrange property elements to
        the orthologGroup elements in the xml. It will add all the levels
        defined in the 'tax'-Taxonomy between the parents most specific
        level and the current nodes level. If no parent exists, all
        tax-levels above the current one are used."""
        self.tax = tax

        top_level_groups = self.parser.getToplevelGroups()

        if PROGRESSBAR and verbosity > 0:
            pbar = setup_progressbar(
                'Adding missing taxonomy annotation: ',
                len(top_level_groups)
            )
            pbar.start()

        for i, fam in enumerate(top_level_groups, start=1):
            try:
                self._addTaxRangeR(fam, noUpwardLevels=not propagate_top)
            except Exception as e:
                logger.exception("cannot process family {}".format(fam.get('id')))
            if PROGRESSBAR and verbosity > 0:
                pbar.update(i)

        if PROGRESSBAR and verbosity > 0:
            pbar.finish()

        del self.tax

    def annotateDoc(self):
        """apply the LOFT naming schema to all the orthologGroups."""
        for i, fam in enumerate(self.parser.getToplevelGroups()):
            self.dupCnt = list()
            self._annotateGroupR(fam, fam.get('id', str(i)))

    def annotateSingletons(self, verbosity=0):
        """Any input genes that aren't assigned to ortholog groups are
        singletons, which are added to the xml as extra ortholog groups"""
        if PROGRESSBAR and verbosity > 0:
            pbar = setup_progressbar('Adding singletons: ', 1)
            pbar.start()

        highest_group = max(self.parser.getToplevelGroups(),
                            key=lambda x: int(
                                x.get('id')))  # TODO: If top-level orthologNodes have no id field this errors out
        input_genes = set(n.get('id') for n in  # Maybe add code to enumerate OG nodes if ids are missing?
                          OrthoXMLQuery.getInputGenes(self.parser.root))
        grouped_genes = set(n.get('id') for n in
                            OrthoXMLQuery.getGroupedGenes(self.parser.root))
        singletons = input_genes - grouped_genes
        groups_node = OrthoXMLQuery.getSubNodes('groups', self.parser.root)[0]

        fam_num = int(highest_group.get('id')) + 1
        singleton_families = set()

        if PROGRESSBAR and verbosity > 0:
            pbar.maxval = len(singletons)

        for i, gene in enumerate(sorted(singletons), start=1):
            singleton_families.add(str(fam_num))
            species = self.parser.mapGeneToSpecies(gene)
            new_node = etree.Element('{{{ns0}}}orthologGroup'.format(
                **self.parser.ns),
                id=str(fam_num))
            new_node.extend(self._createTaxRangeTags(species))
            new_node.append(etree.Element('{{{ns0}}}geneRef'.format(
                **self.parser.ns),
                id=gene))
            groups_node.append(new_node)
            fam_num += 1
            if PROGRESSBAR and verbosity > 0:
                pbar.update(i)

        self.parser.singletons = singleton_families
        if PROGRESSBAR and verbosity > 0:
            pbar.finish()
